/**
 * kevinGD.com Inc. Copyright (c) 2015-2016-2016 All Rights Reserved.
 */
package com.jqtx.windows.utils.http;

import org.apache.commons.lang3.concurrent.BasicThreadFactory;
import org.apache.http.HttpEntityEnclosingRequest;
import org.apache.http.HttpRequest;
import org.apache.http.NameValuePair;
import org.apache.http.client.HttpRequestRetryHandler;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.client.entity.UrlEncodedFormEntity;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.client.protocol.HttpClientContext;
import org.apache.http.config.Registry;
import org.apache.http.config.RegistryBuilder;
import org.apache.http.conn.ConnectTimeoutException;
import org.apache.http.conn.socket.ConnectionSocketFactory;
import org.apache.http.conn.socket.LayeredConnectionSocketFactory;
import org.apache.http.conn.socket.PlainConnectionSocketFactory;
import org.apache.http.conn.ssl.SSLConnectionSocketFactory;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.impl.client.HttpClients;
import org.apache.http.impl.conn.PoolingHttpClientConnectionManager;
import org.apache.http.message.BasicNameValuePair;
import org.apache.http.protocol.HttpContext;
import org.apache.http.util.EntityUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.net.ssl.SSLException;
import javax.net.ssl.SSLHandshakeException;
import java.io.IOException;
import java.io.InterruptedIOException;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.UnknownHostException;
import java.util.*;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * 为对外发http请求连接可复用作处理
 * 采用PoolingHttpClientConnectionManager来管理client对象
 *
 * 详情参见{@link http://www.yeetrack.com/?p=782}
 *
 * @author jiadong.yejd
 * @version $Id: HttpClientPools.java, v 0.1 2016年10月27日 下午5:17:31 jiadong.yejd Exp $
 */
public class HttpClientPools {

  /**logger*/
  private static final Logger logger = LoggerFactory
      .getLogger(HttpClientPools.class);

  private static PoolingHttpClientConnectionManager cm = null;

  /**
   * ConnectTimeout服务器请求超时
   * ConnectionRequestTimeout 建立请求超时
   * SocketTimeout 服务器响应超时
   *
   */
  protected static RequestConfig requestConfig = RequestConfig
      .custom()
      .setConnectTimeout(5000)
      .setConnectionRequestTimeout(
          3000)
      .setSocketTimeout(8000)
      .build();

  static {
    ConnectionSocketFactory plainsf = PlainConnectionSocketFactory.getSocketFactory();
    LayeredConnectionSocketFactory sslsf = SSLConnectionSocketFactory.getSocketFactory();
    Registry<ConnectionSocketFactory> registry = RegistryBuilder.<ConnectionSocketFactory>

        create().

        register("http", plainsf).

        register("https", sslsf).

        build();

    cm = new PoolingHttpClientConnectionManager(registry);

    // 将最大连接数增加到200
    cm.setMaxTotal(200);
    // 将每个路由基础的连接增加到200
    cm.setDefaultMaxPerRoute(200);

    //连接池管理
    IdleConnMonitorTask idleConnMonitorTask = new IdleConnMonitorTask();
    //org.apache.commons.lang3.concurrent.BasicThreadFactory
    ScheduledExecutorService executorService = new ScheduledThreadPoolExecutor(1,
        new BasicThreadFactory.Builder().namingPattern("example-schedule-pool-%d").daemon(true).build());
    executorService.scheduleAtFixedRate(idleConnMonitorTask,30,60, TimeUnit.SECONDS);
    //Timer timer = new Timer(true);
    ////1分钟执行一次
    //timer.schedule(idleConnMonitorTask, 30 * 1000, 60 * 1000);

  }

  /**
   * 它管理着连接池，可以同时为很多线程提供http连接请求
   * 如果连接池有有可用的持久连接，连接管理器就会使用其中的一个，而不是再创建一个新的连接。
   * PoolingClientConnectionManager会根据它的配置来分配请求连接,如果连接池中的所有连接都被占用了，那么后续的请求就会被阻塞，直到有连接被释放回连接池中
   * @return
   */
  public static CloseableHttpClient getHttpClient() {

    //请求重试处理
    HttpRequestRetryHandler httpRequestRetryHandler = new RetryHandler();

    CloseableHttpClient httpClient = HttpClients.custom().setConnectionManager(cm)
        .setRetryHandler(httpRequestRetryHandler).build();
    return httpClient;
  }

  /**
   * 构建Get请求方法
   *
   * @param url 请求地址
   * @param paramMap 参数
   * @param charSet 编码
   * @return
   * @throws Exception
   */
  public static HttpGet buildGet(String url, Map<String, String> paramMap, String charSet)
      throws Exception {
    HttpGet httpGet = new HttpGet(url);

    httpGet.setConfig(requestConfig);

    List<NameValuePair> params = new ArrayList<NameValuePair>();
    Set<String> keySet = paramMap.keySet();
    for (Iterator<String> it = keySet.iterator(); it.hasNext(); ) {
      String next = it.next();
      params.add(new BasicNameValuePair(next, paramMap.get(next)));
    }

    String str = EntityUtils.toString(new UrlEncodedFormEntity(params, charSet));

    httpGet.setURI(new URI(url + "?" + str));

    return httpGet;
  }

  /**
   * 构建Post请求方法
   *
   * @param url 请求地址
   * @param paramMap 参数
   * @param charSet 编码
   * @return
   * @throws UnsupportedEncodingException
   */
  public static HttpPost buildPost(String url, Map<String, String> paramMap, String charSet)
      throws UnsupportedEncodingException {
    HttpPost httpPost = new HttpPost(url);
    httpPost.setConfig(requestConfig);
    httpPost.addHeader("Content-Type", "application/x-www-form-urlencoded;");
    httpPost.addHeader("charset=", charSet);

    List<NameValuePair> valuePairs = new ArrayList<NameValuePair>();
    for (String key : paramMap.keySet()) {
      valuePairs.add(new BasicNameValuePair(key, (String) paramMap.get(key)));
    }

    httpPost.setEntity(new UrlEncodedFormEntity(valuePairs, charSet));
    return httpPost;

  }

  /**
   * 对失败的处理
   *
   * @author jiadong.yejd
   * @version $Id: HttpClientPools.java, v 0.1 2016年10月27日 下午5:53:25 jiadong.yejd Exp $
   */
  static class RetryHandler implements HttpRequestRetryHandler {

    @Override
    public boolean retryRequest(IOException exception, int executionCount, HttpContext context) {
      if (executionCount >= 3) {// 如果已经重试了3次，就放弃
        return false;
      }
      if (exception instanceof SSLHandshakeException) {// 不要重试SSL握手异常
        return false;
      }
      if (exception instanceof InterruptedIOException) {// 超时
        return false;
      }
      if (exception instanceof UnknownHostException) {// 目标服务器不可达
        return false;
      }
      if (exception instanceof ConnectTimeoutException) {// 连接被拒绝
        return false;
      }
      if (exception instanceof SSLException) {// ssl握手异常
        return false;
      }

      HttpClientContext clientContext = HttpClientContext.adapt(context);
      HttpRequest request = clientContext.getRequest();
      // 如果请求是幂等的，就再次尝试
      if (!(request instanceof HttpEntityEnclosingRequest)) {
        return true;
      }
      return false;
    }
  }

  /**
   *
   * @author liudehong
   * @version $Id: HttpClientPools.java, v 0.1 2017年7月31日 下午4:48:48 liudehong Exp $
   */
  private static class IdleConnMonitorTask extends TimerTask {

    public IdleConnMonitorTask() {
      System.err.println("IdleConnMonitorTask...");
    }

    @Override
    public void run() {
      try {
        // 关闭过期的连接
        //cm.closeExpiredConnections();
        //关闭3分钟内不活动的连接
        cm.closeIdleConnections(3, TimeUnit.MINUTES);
        logger.warn("http pool连接池回收成功,连接池的当前状态:" + cm.getTotalStats());
      } catch (Throwable ex) {
        logger.error("http连接池管理异常", ex);
      }
    }
  }

}
